import os
import leveldb
import lib_compute.typon_math as typon_math
import lib_dsa.byte_array as byte_array
import lib_dsa.kv_iter as kv_iter

class kv_iterator_vba():
    
    def __init__(self, lvl_db_iter, kvs_ref, include_key, include_value):
        self.lvl_db_iter = lvl_db_iter
        self.include_key = include_key
        self.include_value = include_value
    
    
    def has_next(self):
        return self.lvl_db_iter.has_next()
    
    
    def get_next(self):
        
        if not self.has_next():
            return None
        
        key, index_repr = self.lvl_db_iter.get_next()
        result = []
        
        if self.include_key:
            result.append(key)
        
        
        if self.include_value:
            item_index = typon_math.deserialize_int( index_repr )
            result.append( self.kvs_ref.vba.select_id( item_index ) )
        
        return result


class kv_iterator_no_vba():
    
    def __init__(self, lvl_db_iter, kvs_ref):
        self.lvl_db_iter = lvl_db_iter
    
    
    def has_next(self):
        return self.lvl_db_iter.has_next()
    
    
    def get_next(self):
        return self.lvl_db_iter.get_next()


class kv_store():
    
    def __init__(self, db_path, vba_engine=False):
        self.db_path = db_path
        self.db = leveldb.LevelDB( db_path, create_if_missing=True )
        self.vba_engine = vba_engine
        
        if vba_engine:
            self.vba = byte_array.make_vba( db_path )
    
    
    def engine_get(self, key):
        
        if not self.vba_engine:
            return self.db.Get(key)
        
        index_repr = self.db.Get(key)
        item_index = typon_math.deserialize_int(index_repr)
        return self.vba.select_id(item_index)
    
    
    def engine_set(self, key, value):
        
        if not self.vba_engine:
            return self.db.Put(key, value)
        
        index_repr = typon_math.serialize_int( self.vba.item_count )
        kvs_op = self.db.Put(key, index_repr)
        le_op = self.vba.insert_bytes( value )
        return [ kvs_op, le_op ]
    
    
    def engine_remove(self, key):
        return self.db.Delete(key)
    
    
    def get(self, key, key_encoding='latin-1'):
        
        if isinstance(key, str):
            key_bytes = key.encode(key_encoding)
            return self.engine_get(key_bytes)
        
        return self.engine_get(key)
    
    
    def set(self, key, value, key_encoding='latin-1', value_encoding='latin-1'):
        
        if isinstance(key, str):
            key_bytes = key.encode(key_encoding)
            
            if isinstance(value, str):
                value_bytes = value.encode(value_encoding)
                return self.engine_set(key_bytes, value_bytes)
            
            return self.engine_set(key_bytes, value)
        
        
        if isinstance(value, str):
            value_bytes = value.encode(value_encoding)
            return self.engine_set(key, value_bytes)
        
        return self.engine_set(key, value)
    
    
    def remove(self, key, key_encoding='latin-1'):
        
        if isinstance(key, str):
            key_bytes = key.encode(key_encoding)
            return self.engine_remove(key_bytes)
        
        return self.engine_remove(key)
    
    
    def batch_write(self, op_args_L1, transaction=False):
        
        if self.vba_engine:
            wb = self.db.write_batch(transaction=transaction)
            value_insert_L1 = []
            insert_count = 0
            
            for op_args in op_args_L1:
                
                if op_args[0] == 'set':
                    index_repr = typon_math.serialize_int( self.vba.item_count + insert_count )
                    insert_count += 1
                    value_insert_L1.append( op_args[2] )
                    wb.put( op_args[1], index_repr )
                    continue
                
                
                if op_args[0] == 'remove':
                    wb.remove( op_args[1] )
            
            kvs_op = wb.write()
            vba_op = self.vba.batch_insert_bytes(value_insert_L1)
            return [ kvs_op, vba_op ]
        
        wb = self.db.write_batch( transaction=transaction )
        
        for op_args in op_args_L1:
            
            if op_args[0] == 'set':
                wb.put( op_args[1], op_args[2] )
                continue
            
            
            if op_args[0] == 'remove':
                wb.remove( op_args[1] )
        
        kvs_op = wb.write()
        return kvs_op
    
    
    def iterator(
            self,
            start=None,
            stop=None,
            prefix=None,
            include_start=True,
            include_stop=False,
            include_key=True,
            include_value=True,
            reverse=False
        ):
        
        if not self.vba_engine:
            lvl_db_iter = kv_iter.ldb_iterator(
                            self.db,
                            start=start,
                            stop=stop,
                            prefix=prefix,
                            include_start=include_start,
                            include_stop=include_stop,
                            include_key=include_key,
                            include_value=include_value,
                            reverse=reverse
                        )
            return kv_iterator_no_vba(lvl_db_iter, self)
        
        lvl_db_iter = kv_iter.ldb_iterator(
                    self.db,
                    start=start,
                    stop=stop,
                    prefix=prefix,
                    include_start=include_start,
                    include_stop=include_stop,
                    include_key=True,
                    include_value=True,
                    reverse=reverse
                )
        return kv_iterator_vba(lvl_db_iter, self, include_key, include_value)



import os
import lib_compute.typon_math as typon_math
import threading

def make_vba(base_path):
    index_fspath = f'{base_path}.index'
    data_fspath = f'{base_path}.data'
    clear_fspath = False
    return variable_bytes_array(index_fspath, data_fspath, clear_fspath)


class variable_bytes_array():
    
    def __init__(self, index_fspath: str, data_fspath: str, clear_fspath: bool):
        self.index_fspath = index_fspath
        self.data_fspath = data_fspath
        self.item_count = 0
        self.total_bytes = 0
        self.size_fxn = os.path.getsize
        self.mutex = threading.Lock()
        
        if clear_fspath:
            fptr = open(self.index_fspath, 'wb')
            fptr.close()
            fptr = open(self.data_fspath, 'wb')
            fptr.close()
            return
        
        
        if not os.path.exists(self.index_fspath):
            open(self.index_fspath, 'wb').close()
        
        
        if not os.path.exists(self.data_fspath):
            open(self.data_fspath, 'wb').close()
        
        self.item_count = self.size_fxn(self.index_fspath) // 16
        self.total_bytes = self.size_fxn(self.data_fspath)
    
    
    def index_size(self) -> int:
        return self.size_fxn(self.index_fspath)
    
    
    def data_size(self) -> int:
        return self.size_fxn(self.data_fspath)
    
    
    def insert_bytes(self, target_bytes) -> int:
        size_repr = typon_math.serialize_int( len(target_bytes) )
        offset_repr = typon_math.serialize_int( self.total_bytes )
        index_bytes = offset_repr + size_repr
        fptr = open(self.index_fspath, 'ab')
        fptr.write(index_bytes)
        fptr.close()
        fptr1 = open(self.data_fspath, 'ab')
        fptr1.write(target_bytes)
        fptr1.close()
        self.total_bytes += len(target_bytes)
        self.item_count += 1
        return self.item_count - 1
    
    
    def batch_insert_bytes(self, target_bytes_L1) -> int:
        current_offset = 0
        fptr = open(self.index_fspath, 'ab')
        fptr1 = open(self.data_fspath, 'ab')
        
        for i in range(0, len(target_bytes_L1)):
            size_repr = typon_math.serialize_int( len(target_bytes_L1[i]) )
            offset_repr = typon_math.serialize_int(self.total_bytes + current_offset)
            fptr.write(offset_repr)
            fptr.write(size_repr)
            fptr1.write( target_bytes_L1[i] )
            current_offset += len(target_bytes_L1[i])
        
        fptr.close()
        fptr1.close()
        self.total_bytes += current_offset
        self.item_count += len(target_bytes_L1)
        return self.item_count - 1
    
    
    def read_index_bytes(self, offset: int, limit: int) -> str:
        fptr = open(self.index_fspath, 'rb')
        fptr.seek(offset, 0)
        result = fptr.read(limit)
        fptr.close()
        return result
    
    
    def read_data_bytes(self, offset: int, limit: int) -> str:
        fptr = open(self.data_fspath, 'rb')
        fptr.seek(offset, 0)
        result = fptr.read(limit)
        fptr.close()
        return result
    
    
    def select_id(self, item_id: int) -> str:
        
        if item_id < 0 or item_id >= self.item_count:
            return ''
        
        fptr = open(self.index_fspath, 'rb')
        fptr.seek(16 * item_id, 0)
        index_repr = fptr.read(16)
        fptr.close()
        item_offset = typon_math.deserialize_int(index_repr[ 0 : 8 ])
        item_size = typon_math.deserialize_int(index_repr[ 8 : 16 ])
        fptr = open(self.data_fspath, 'rb')
        fptr.seek(item_offset, 0)
        result = fptr.read(item_size)
        fptr.close()
        return result
    
    
    def select_id_range(self, id_lower_bound: int, id_upper_bound: int):
        
        if id_lower_bound < 0 or id_upper_bound > self.item_count or id_lower_bound >= id_upper_bound:
            return []
        
        fptr = open(self.index_fspath, 'rb')
        fptr.seek(16 * id_lower_bound, 0)
        index_repr = fptr.read(16 * (id_upper_bound - id_lower_bound))
        fptr.close()
        fptr = open(self.data_fspath, 'rb')
        result: vec[str] = []
        
        for i in range(id_lower_bound, id_upper_bound):
            index_offset = 16 * (i - id_lower_bound)
            item_offset = typon_math.deserialize_int(index_repr[ index_offset + 0 : index_offset + 8 ])
            item_size = typon_math.deserialize_int(index_repr[ index_offset + 8 : index_offset + 16 ])
            fptr.seek(item_offset, 0)
            item_repr = fptr.read(item_size)
            result.append(item_repr)
        
        fptr.close()
        return result
    
    
    def remove_last_k(self, k: int) -> int:
        
        if k <= 0 or k > self.item_count:
            return 0
        
        item_id = self.item_count - k
        fptr = open(self.index_fspath, 'rb+')
        fptr.seek(16 * item_id, 0)
        index_repr = fptr.read(16)
        item_offset = typon_math.deserialize_int(index_repr[ 0 : 8 ])
        fptr.truncate(16 * item_id)
        fptr.close()
        fptr1 = open(self.data_fspath, 'rb+')
        fptr1.truncate(item_offset)
        fptr1.close()
        self.item_count -= k
        return item_id



import struct

def deserialize_int(int_val_repr):
    return int.from_bytes(int_val_repr, 'little', signed=True)


def deserialize_int32(int_val_repr):
    return int.from_bytes(int_val_repr, 'little', signed=True)


def serialize_int(int_val_arg):
    return int_val_arg.to_bytes(8, byteorder='little', signed=True)


def serialize_int32(int_val_arg):
    return int_val_arg.to_bytes(4, byteorder='little', signed=True)


def deserialize_dbl(dbl_val_repr):
    result = struct.unpack('<d', dbl_val_repr)[0]
    return result


def serialize_dbl(dbl_val_arg):
    result = struct.pack('<d', dbl_val_arg)
    return result



import lib_py_parse.utils.constants as constants
import lib_py_parse.utils.exceptions as exceptions
import lib_py_parse.utils.parsing_utils as parsing_utils
import lib_py_parse.helper.type_resolver as type_resolver

def extract_id_var_tuple(S2):
    L1 = [x.strip() for x in S2.split(':')]
    
    if len(L1) != 2:
        error_msgs = [
                    'invalid fxn argument declaration',
                    'arguments must be formatted as:',
                    '({arg_1}: {type_1}, ..., {arg_i}: {type_i}, ..., {arg_n}: {type_n})',
                    f'input: {S2}',
                ]
        exceptions.raise_exception_ue(error_msgs)
    
    return L1


def read_fxn_params(arg_str):
    S1 = arg_str[1:-1]
    
    if len(S1) == 0:
        return [], []
    
    fparam_ids = []
    arg_types_typon = []
    i1 = 0
    index = 0
    
    while index < len(S1):
        
        if S1[index] == '[':
            index = parsing_utils.get_char_closure(S1, index+1)
            continue
        
        
        if S1[index] == ',':
            S2 = S1[i1:index]
            fparam_id, arg_type_typon = extract_id_var_tuple(S2)
            fparam_ids.append(fparam_id)
            arg_types_typon.append(arg_type_typon)
            i1 = index + 1
        
        index += 1
    
    
    if i1 < index:
        S2 = S1[i1:index]
        fparam_id, arg_type_typon = extract_id_var_tuple(S2)
        fparam_ids.append(fparam_id)
        arg_types_typon.append(arg_type_typon)
    
    return fparam_ids, arg_types_typon


def fmt_type_module_space_aware_module_L0_skeleton(input_typon_type, current_module, module_L0_skeleton):
    S1 = input_typon_type.replace(' ', '')
    c_ext_type_prefix = constants.select_item_prefix('c_ext_type_prefix')
    primitives_map = constants.select_primitives()
    
    if S1 in primitives_map:
        return S1
    
    
    if S1[0:3] == c_ext_type_prefix:
        return S1
    
    
    if S1[0:4] == 'vec[':
        item_type = S1[4:-1]
        item_type_msa = fmt_type_module_space_aware_module_L0_skeleton(item_type, current_module, module_L0_skeleton)
        return f'vec[{item_type_msa}]'
    
    
    if S1[0:5] == 'hmap[':
        key_type, value_type = type_resolver.resolve_map_type_L0_skeleton(S1, module_L0_skeleton)
        key_type_msa = fmt_type_module_space_aware_module_L0_skeleton(key_type, current_module, module_L0_skeleton)
        value_type_msa = fmt_type_module_space_aware_module_L0_skeleton(value_type, current_module, module_L0_skeleton)
        return f'hmap[{key_type_msa}, {value_type_msa}]'
    
    
    if S1[0:5] == 'smap[':
        key_type, value_type = type_resolver.resolve_map_type_L0_skeleton(S1, module_L0_skeleton)
        key_type_msa = fmt_type_module_space_aware_module_L0_skeleton(key_type, current_module, module_L0_skeleton)
        value_type_msa = fmt_type_module_space_aware_module_L0_skeleton(value_type, current_module, module_L0_skeleton)
        return f'smap[{key_type_msa}, {value_type_msa}]'
    
    
    if S1[0:5] == 'tupl[':
        typon_tuple_typon_type_list = type_resolver.read_tuple_types_typon_L0_skeleton(S1, module_L0_skeleton)
        L1 = [
                    fmt_type_module_space_aware_module_L0_skeleton(T1, current_module, module_L0_skeleton)
                    for T1 in typon_tuple_typon_type_list
                ]
        S2 = ', '.join(L1)
        return f'tupl[{S2}]'
    
    
    if S1[0:4] == 'fxn[':
        arg_types_typon, return_type_typon = type_resolver.read_args_ret_from_fxn_typon_type(S1)
        L1 = [
                    fmt_type_module_space_aware_module_L0_skeleton(T1, current_module, module_L0_skeleton)
                    for T1 in arg_types_typon
                ]
        return_type_typon_msa = fmt_type_module_space_aware_module_L0_skeleton(return_type_typon, current_module, module_L0_skeleton)
        S2 = ', '.join(L1)
        return f'fxn[[{S2}], {return_type_typon_msa}]'
    
    user_defined_typon_class_type = S1
    
    if user_defined_typon_class_type.find('|') != -1:
        return f'{user_defined_typon_class_type}'
    
    i1 = user_defined_typon_class_type.find('.')
    
    if i1 == -1:
        return f'{current_module}|{user_defined_typon_class_type}'
    
    L1 = user_defined_typon_class_type.split('.')
    assert len(L1) == 2
    module_alias, class_name = L1
    fully_qualified_module_name = module_L0_skeleton['imports']['alias_map'][module_alias]
    return f'{fully_qualified_module_name}|{class_name}'



import lib_py_parse.utils.exceptions as exceptions
import lib_py_parse.utils.constants as constants

def indent_multiline_str(indent_str, multiline_str):
    return '\n'.join(
            [
                f'{indent_str}{x}'
                for x in multiline_str.splitlines()
            ]
        )


def get_fmt_str_expr_closure(target_substr, open_index, symbol_table):
    assert target_substr[open_index - 1] == '{'
    op_count = 1
    index = open_index
    
    while op_count > 0 and index < len(target_substr):
        
        if target_substr[index] == '}':
            op_count -= 1
        
        
        if target_substr[index] == '{':
            op_count += 1
        
        
        if target_substr[index] in ['\'', '\"']:
            index = get_quote_closure(target_substr, index)
            continue
        
        index += 1
    
    
    if op_count > 0:
        err_message = f'str literal cannot be parsed: {target_substr}'
        exceptions.raise_exception_ST(symbol_table, err_message)
    
    return index


def get_char_closure(s, index):
    t_mapper = constants.select_bracket_map()
    i = index
    open_token, close_token = s[index-1], t_mapper[s[index-1]]
    scope = 1
    
    while i < len(s) and scope > 0:
        
        if s[i] == open_token:
        

scope += 1

if s[i] == close_token:


scope -= 1
i += 1

if scope > 0:
    exceptions.raise_exception_msg(s, index, f'Unclosed: {s[index-1]}')
    return i


def get_unescaped_quote(s, index, quote):
    assert s[index] == quote
    i1 = index
    
    while True:
        i1 = s.find(quote, i1+1)
        
        if i1 == -1:
            exceptions.raise_exception_msg(s, index, f'unclosed quote')
        
        esc_count = 0
        
        while esc_count < i1 and s[i1 - esc_count - 1] == '\\':
            esc_count += 1
        
        
        if esc_count % 2 == 0:
            break
    
    return i1


def close_triple_quote(s, index, trq):
    i1 = index-1
    
    while True:
        i1 = s.find(trq, i1+1)
        
        if i1 == 0:
            break
        
        esc_count = 0
        
        while esc_count < i1 and s[i1 - esc_count - 1] == '\\':
            esc_count += 1
        
        
        if esc_count % 2 == 0:
            break
        
        
        if i1 == -1:
            raise Exception('unclosed triple quote')
    
    return i1


def get_quote_closure(s, index):
    trq1 = '\'\'\''
    trq2 = '\'\'\''
    
    if s[index:index+3] == trq1:
        return close_triple_quote(s, index+3, trq1) + 3
    
    
    if s[index:index+3] == trq2:
        return close_triple_quote(s, index+3, trq2) + 3
    
    return get_unescaped_quote(s, index, s[index]) + 1


def is_escaped_quote(s, index):
    result = 0
    
    while index > result and s[index - result - 1] == '\\':
        result += 1
    
    return result % 2 == 1


def get_cleared_char_closure(s, index):
    t_mapper = constants.select_bracket_map()
    i = index
    open_token, close_token = s[index-1], t_mapper[s[index-1]]
    scope = 1
    
    while i < len(s) and scope > 0:
        
        if s[i] == '\'' and not is_escaped_quote(s, i):
            i = get_quote_closure(s, i)
            continue
        
        
        if s[i] == '\"' and not is_escaped_quote(s, i):
            i = get_quote_closure(s, i)
            continue
        
        
        if s[i] == open_token:
        

scope += 1

if s[i] == close_token:


scope -= 1
i += 1

if scope > 0:
    exceptions.raise_exception_msg(s, index, f'Unclosed: {s[index-1]}')
    return i


def read_next_cleared_char(s, index, target_char):
    t_mapper = constants.select_bracket_map()
    
    while index < len(s):
        
        if s[index] == target_char:
            break
        
        
        if s[index] in t_mapper:
            index = get_cleared_char_closure(s, index+1)
        
        
        elif s[index] == '\'' and not is_escaped_quote(s, index):
            index = get_quote_closure(s, index)
        
        
        elif s[index] == '\"' and not is_escaped_quote(s, index):
            index = get_quote_closure(s, index)
        
        
        else :
            index += 1
    
    
    if index == len(s):
        return -1
    
    return index


def parse_current_scope(source_code, scope_start_index):
    current_scope = 0
    index = scope_start_index
    
    if not source_code[index].isspace():
        return current_scope, index
    
    scope_divide_factor_router = {
            '\t' : 1,
            ' ' : 4,
        }
    scope_divide_factor = scope_divide_factor_router[source_code[index]]
    
    while index < len(source_code) and source_code[index].isspace():
        index += 1
        current_scope += 1
    
    current_scope = current_scope // scope_divide_factor
    return current_scope, index


def read_init_token(source_code, index, scope):
    
    while index < len(source_code) and source_code[index] in ['\r', '\n']:
        index += 1
    
    
    if index == len(source_code):
        return 'EOF', index
    
    line_begin = index
    current_scope, index = parse_current_scope(source_code, index)
    
    if index == len(source_code):
        return 'EOF', index
    
    
    if source_code[index] == '#':
        return source_code[index], index+1
    
    
    if scope > current_scope:
        return 'EOS', line_begin
    
    i_start = index
    
    while index < len(source_code):
        
        if source_code[index].isspace() or source_code[index] == ':':
            break
        
        index += 1
    
    result = source_code[i_start:index]
    return result, index


def select_token_index(tok_stream, target_token):
    index = -1
    
    for i in range(0, len(tok_stream)):
        
        if tok_stream[i] == target_token:
            index = i
            break
    
    return index


def get_token_stream_closure(tok_stream, i_start):
    t_mapper = constants.select_bracket_map()
    index = -1
    bound = 0
    open_token = tok_stream[index]
    close_token = t_mapper[tok_stream[index]]
    
    for i in range(i_start, len(tok_stream)):
        
        if tok_stream[i] == close_token:
            bound -= 1
        
        
        if tok_stream[i] == open_token:
            bound += 1
        
        
        if bound == 0:
            index = i
            break
    
    return index


def get_expr_aware_closure(tok_stream, index, symbol_table):
    mapper = {
            't[' : ']',
            '[' : ']',
            '(' : ')',
            '{' : '}',
            's{' : '}',
        }
    assert tok_stream[index] in mapper
    target_token = mapper[tok_stream[index]]
    i1 = index+1
    
    while i1 < len(tok_stream):
        
        if tok_stream[i1] == target_token:
            i1 += 1
            break
        
        
        if tok_stream[i1] in mapper:
            i1 = get_expr_aware_closure(tok_stream, i1, symbol_table)
            continue
        
        i1 += 1
    
    
    if tok_stream[i1-1] != target_token:
        err_message = f'could not resolve expr_aware_closure: {tok_stream} {index}'
        exceptions.raise_exception_ST(symbol_table, err_message)
    
    return i1


def find_all_expr_aware(tok_stream, index, symbol_table, target_token):
    mapper = {
            't[' : ']',
            '[' : ']',
            '(' : ')',
            '{' : '}',
            's{' : '}',
        }
    result = []
    i1 = index
    
    while i1 < len(tok_stream):
        
        if tok_stream[i1] == target_token:
            result.append(i1)
        
        
        if tok_stream[i1] in mapper:
            i1 = get_expr_aware_closure(tok_stream, i1, symbol_table)
            continue
        
        i1 += 1
    
    return result


def split_expr_aware(tok_stream, symbol_table, target_token):
    indices = find_all_expr_aware(tok_stream, 0, symbol_table, target_token)
    last_index = -1
    result = []
    substream = []
    
    for i in range(0, len(indices)):
        substream = tok_stream[last_index+1:indices[i]]
        last_index = indices[i]
        result.append(substream)
    
    substream = tok_stream[last_index+1:]
    result.append(substream)
    return result


def split_arg_typle(s, index):
    i = index
    i0 = index
    result = []
    
    while i < len(s):
        
        if s[i] == '[':
            i = get_char_closure(s, i+1) + 1
            continue
        
        
        if s[i] == ',':
            result.append(s[i0:i])
            i0 = i
        
        i += 1
    
    
    if i0 < len(s):
        result.append(s[i0:])
    
    return result


def read_delimited_types(type_str):
    result = []
    index = 0
    i1 = 0
    
    while index < len(type_str):
        
        if type_str[index] == ',':
            result.append(type_str[i1:index])
            ctoken = ''
            index += 1
            
            while index < len(type_str) and type_str[index].isspace():
                index += 1
            
            i1 = index
            continue
        
        
        if type_str[index] == '[':
            index = get_char_closure(type_str, index+1)
            continue
        
        index += 1
    
    
    if index - i1 > 0:
        result.append(type_str[i1:index])
    
    return result


def parse_fxn_type(function_type, symbol_table):
    ws_removed_typon_fxn_type = function_type.replace(' ', '')
    interior_fxn_type = ws_removed_typon_fxn_type[4:-1]
    
    if len(interior_fxn_type) == 0:
        error_msgs = [
                    'invalid type declaration',
                    f'empty fxn not allowed: {ws_removed_typon_fxn_type}',
                ]
        exceptions.raise_exception_ST(symbol_table, error_msgs)
    
    arg_closure_index = get_char_closure(interior_fxn_type, 1)
    typon_fxn_arg_str = interior_fxn_type[1:arg_closure_index-1]
    arg_types_typon = read_delimited_types(typon_fxn_arg_str)
    return_type_typon = interior_fxn_type[arg_closure_index+1:]
    return arg_types_typon, return_type_typon



import lib_py_parse.helper.eval_helper as eval_helper
import lib_py_parse.utils.exceptions as exceptions

def resolve_func_type_from_arguments(arguments):
    T1 = [
            T2.strip()
            for T2 in arguments.split('->')
        ]
    
    if len(T1) != 2:
        error_msgs = [
                    'invalid fxn statement',
                    'fxn statement must be formatted as:',
                    'fxn ({arguments}) -> {return_type}:',
                    f'input: {arguments}',
                ]
        exceptions.raise_exception_ue(error_msgs)
    
    fparam_ids, arg_types_typon = eval_helper.read_fxn_params(T1[0])
    return_type_typon = T1[1]
    return fparam_ids, arg_types_typon, return_type_typon


def resolve_args_only_func_type(arguments):
    T1 = [
            T2.strip()
            for T2 in arguments.split('->')
        ]
    
    if len(T1) != 1:
        error_msgs = [
                    'invalid fxn argument declaration',
                    'arguments must be formatted as:',
                    '({arg_1}: {type_1}, ..., {arg_i}: {type_i}, ..., {arg_n}: {type_n})',
                    f'input: {arguments}',
                ]
        exceptions.raise_exception_ue(error_msgs)
    
    fparam_ids, arg_types_typon = eval_helper.read_fxn_params(T1[0])
    return fparam_ids, arg_types_typon


def resolve_func_type_from_arguments_msa(arguments, module_L0_skeleton, current_module):
    fparam_ids, arg_types_typon, return_type_typon = resolve_func_type_from_arguments(arguments)
    arg_types_typon_msa = [
            eval_helper.fmt_type_module_space_aware_module_L0_skeleton(arg_type_typon, current_module, module_L0_skeleton)
            for arg_type_typon in arg_types_typon
        ]
    return_type_typon_msa = eval_helper.fmt_type_module_space_aware_module_L0_skeleton(return_type_typon, current_module, module_L0_skeleton)
    return fparam_ids, arg_types_typon_msa, return_type_typon_msa


def resolve_args_only_func_type_msa(arguments, module_L0_skeleton, current_module):
    fparam_ids, arg_types_typon = resolve_args_only_func_type(arguments)
    arg_types_typon_msa = [
            eval_helper.fmt_type_module_space_aware_module_L0_skeleton(arg_type_typon, current_module, module_L0_skeleton)
            for arg_type_typon in arg_types_typon
        ]
    return fparam_ids, arg_types_typon_msa


def is_valid_var_name(var_name):
    
    for i in range(0, len(var_name)):
        
        if not var_name[i].isalnum() and var_name[i] != '_':
            return False
    
    return True



import os, json
import lib_py_parse.utils.exceptions as exceptions

def format_indent(indent_str, multiline_str):
    return '\n'.join([f'{indent_str}{x}' for x in multiline_str.splitlines()])


def serialize_json(py_dict):
    result = json.dumps(py_dict, indent=1)
    return result


def print_json(py_dict):
    result = json.dumps(py_dict, indent=4)
    print(result)
    return 1


def select_module_src_dir(module_dir, target_module):
    x1 = target_module.replace('.', '/')
    result = f'{module_dir}/{x1}.py'
    
    if not os.path.exists(result):
        error_msgs = [
                    'module does not exist:',
                    f'module: {target_module}',
                ]
        exceptions.raise_exception_ue(error_msgs)
    
    return result


def read_module_src(module_dir, target_module):
    x1 = target_module.replace('.', '/')
    result = f'{module_dir}/{x1}.py'
    
    if not os.path.exists(result):
        error_msgs = [
                    'module does not exist:',
                    f'module: {target_module}',
                ]
        exceptions.raise_exception_ue(error_msgs)
    
    module_src = open(result, 'r').read()
    return result



import os

def perform_dfs(graph, i, v, cv, result_order, cyclic_flag):
    cv[i] = 1
    
    for x in graph[i]:
        
        if x in v:
            continue
        
        
        if x in cv:
            cyclic_flag[0] = True
            return
        
        perform_dfs(graph, x, v, cv, result_order, cyclic_flag)
    
    result_order.append(i)
    del cv[i]
    v[i] = 1


def top_sort(graph):
    result_order = []
    cyclic_flag = {0: False}
    v = {}
    cv = {}
    
    for i in range(0, len(graph)):
        
        if i not in v:
            perform_dfs(graph, i, v, cv, result_order, cyclic_flag)
        
        
        if cyclic_flag[0]:
            error_msgs = [
                            f'module contains cyclic imports',
                            f'cycle beginning from module: {module_names[i]}',
                        ]
            raise Exception( '\n'.join(error_msgs) )
    
    return result_order[::-1]


def bfs_files(base_dir, target_exts):
    target_exts_map = { x : 1 for x in target_exts }
    q = [base_dir]
    result = []
    
    while len(q) > 0:
        fspath = q.pop()
        
        if os.path.isdir(fspath):
            
            for file in os.listdir(fspath):
                q.append( f'{fspath}/{file}' )
            
            continue
        
        i1 = fspath.rfind('.', 0)
        ext = fspath[i1+1 : len(fspath) ]
        
        if ext in target_exts_map or len(target_exts) == 0:
            result.append(fspath)
    
    return result



import random

def partition_batch_size(data, batch_size):
    result = []
    
    if batch_size <= 0:
        return result
    
    k = len(data) // batch_size
    
    for i in range(0, k):
        result.append(data[i * batch_size : (i+1) * batch_size])
    
    
    if len(data) % batch_size > 0:
        result.append(data[k * batch_size : ])
    
    return result


def partition_num_batches(data, num_batches):
    result = []
    
    if num_batches <= 0:
        return result
    
    batch_size = len(data) // num_batches
    r = len(data) % num_batches
    
    for i in range(0, num_batches):
        
        if i < r:
            result.append( data[ i * (batch_size+1): (i+1) * (batch_size + 1)] )
            continue
        
        result.append( data[ r * (batch_size+1) + (i-r) * batch_size : r * (batch_size+1) + (i-r+1) * batch_size ] )
    
    return result


def partition_data(data, batch_size):
    result = []
    
    if batch_size <= 0:
        return result
    
    k = len(data) // batch_size
    
    for i in range(0, k-1):
        result.append(data[i * batch_size : (i+1) * batch_size])
    
    result.append(data[(k-1) * batch_size : ])
    return result



import lib_compute.typon_math as typon_math
import lib_py_parse.utils.parsing_utils as parsing_utils

def serialize_data(target_value, target_type, classes):
    target_type = target_type.replace(' ', '')
    
    if target_type == 'bool':
        result = b'\x00'
        
        if target_value:
            result = b'\x01'
        
        return result
    
    
    if target_type == 'char':
        X1 =  str(target_value)
        
        if len(X1) != 1:
            return b'\x00'
        
        
        try:
            return X1.encode('latin-1')
        
        
        except:
            return X1.encode()
    
    
    if target_type == 'int':
        result = typon_math.serialize_int(target_value)
        return result
    
    
    if target_type == 'dbl':
        result = typon_math.serialize_dbl(target_value)
        return result
    
    
    if target_type == 'str':
        result = None
        
        if isinstance(target_value, bytes):
            result = typon_math.serialize_int( len(target_value) ) + target_value
        
        
        else :
            
            try:
                result = typon_math.serialize_int( len(target_value) ) + target_value.encode('latin-1')
            
            
            except:
                result = typon_math.serialize_int( len(target_value) ) + target_value.encode()
        
        return result
    
    
    if target_type[0:4] == 'vec[':
        result = typon_math.serialize_int( len(target_value) )
        item_type = target_type[4:-1]
        
        for i in range(0, len(target_value)):
            result += serialize_data( target_value[i], item_type, classes )
        
        return result
    
    
    if target_type[0:5] in ['hmap[', 'smap[']:
        delimited_types = target_type[5:-1]
        key_type, value_type = parsing_utils.read_delimited_types(delimited_types)
        result = typon_math.serialize_int( len(target_value) )
        
        for key_var in target_value:
            result += serialize_data( key_var, key_type, classes )
            result += serialize_data( target_value[key_var], value_type, classes )
        
        return result
    
    
    if target_type[0:5] == 'tupl[':
        delimited_types = target_type[5:-1]
        typon_tuple_types = parsing_utils.read_delimited_types(delimited_types)
        result = b''
        
        for i in range(0, len(typon_tuple_types)):
            result += serialize_data( target_value[i], typon_tuple_types[i], classes )
        
        return result
    
    
    if target_type in classes:
        class_items = classes[target_type]
        result = b''
        
        for attr_name in class_items:
            attr_value = eval( f'target_value.{attr_name}' )
            attr_type = class_items[attr_name]
            result += serialize_data( attr_value, attr_type, classes )
        
        return result
    
    raise Exception(f'unverified type in fxn serialize_data: {target_type}')



import setuptools

setuptools.setup(
    name='lib-V4',
    version='0.0.0',
    description='Local packages for development',

    author='hypothesisbase',
    author_email="support@hypothesisbase.com",
    url= f'http://pypi.python.org/pypi/lib-V4/0.0.0',
    license="MIT",

    install_requires=[
        "bs4",
        "leveldb",
        "selenium",
    ],
    packages=setuptools.find_packages(),
    zip_safe=False
)



import os
import lib_compute.typon_math as typon_math
import threading

class threadable_queue():
    
    def __init__(self):
        self.mutex = threading.Lock()
        self.in_buf = []
        self.out_buf = []
    
    
    def size(self):
        with self.mutex:
        return len(self.in_buf) + len(self.out_buf)
    
    
    def push(self, value):
        with self.mutex:
        self.out_buf.append(value)
        return True
    
    
    def poll(self):
        with self.mutex:
        
        if len(self.in_buf) == 0:
            
            while len(self.out_buf) > 0:
                self.in_buf.append( self.out_buf.pop() )
                
                if len(self.in_buf) == 0:
                    return None
            
            return self.in_buf.pop()


class threadable_hmap():
    
    def __init__(self):
        self.mutex = threading.Lock()
        self.data_map = {}
    
    
    def size(self):
        with self.mutex:
        return len(self.data_map)
    
    
    def get(self, key):
        with self.mutex:
        
        if key not in self.data_map:
            return None
            return self.data_map[key]
    
    
    def set(self, key, value):
        with self.mutex:
        self.data_map[key] = value
        return True
    
    
    def remove(self, key):
        with self.mutex:
        
        if key not in self.data_map:
            return False
            del self.data_map[key]
            return True
    
    
    def keys(self):
        with self.mutex:
        return [ x for x in self.data_map ]


class threadable_vec():
    
    def __init__(self):
        self.mutex = threading.Lock()
        self.data_map = []
    
    
    def size(self):
        with self.mutex:
        return len(self.data_map)
    
    
    def get(self, key):
        with self.mutex:
        
        if key >= len(self.data_map) or key < 0:
            return None
            return self.data_map[key]
    
    
    def set(self, key, value):
        with self.mutex:
        
        if key >= len(self.data_map) or key < 0:
            return None
            self.data_map[key] = value
            return True
    
    
    def remove(self, key):
        with self.mutex:
        
        if key >= len(self.data_map) or key < 0:
            return None
            self.data_map[key] = None
            return True
    
    
    def keys(self):
        with self.mutex:
        return [ x for x in self.data_map ]



import lib_py_parse.utils.parsing_utils as parsing_utils
import lib_py_parse.utils.exceptions as exceptions
import lib_py_parse.utils.constants as constants

def reduce_rel_type( target_module, current_module, rel_type, reverse_imports_map ):
    target_rel = f'{rel_type}'
    
    if target_module != current_module:
        import_alias = reverse_imports_map[current_module][target_module]
        target_rel = f'{import_alias}.{rel_type}'
    
    return target_rel


def compute_relative_type(abs_type, current_module, reverse_imports_map):
    S1 = abs_type.replace(' ', '')
    c_ext_type_prefix = constants.select_item_prefix('c_ext_type_prefix')
    primitives_map = constants.select_primitives()
    
    if S1 in primitives_map:
        return current_module, S1
    
    
    if S1[0:3] == c_ext_type_prefix:
        return current_module, S1
    
    
    if S1[0:4] == 'vec[':
        item_type = S1[4:-1]
        item_module, item_rel_type = compute_relative_type(item_type, current_module, reverse_imports_map)
        item_rel = reduce_rel_type( item_module, current_module, item_rel_type, reverse_imports_map )
        return current_module, f'vec[{item_rel}]'
    
    
    if S1[0:5] == 'hmap[':
        key_type, value_type = resolve_map_type_L0_skeleton(S1, reverse_imports_map)
        key_module, key_rel_type = compute_relative_type(key_type, current_module, reverse_imports_map)
        value_module, value_rel_type = compute_relative_type(value_type, current_module, reverse_imports_map)
        key_rel = reduce_rel_type( key_module, current_module, key_rel_type, reverse_imports_map )
        value_rel = reduce_rel_type( value_module, current_module, value_rel_type, reverse_imports_map )
        return current_module, f'hmap[{key_rel}, {value_rel}]'
    
    
    if S1[0:5] == 'smap[':
        key_type, value_type = resolve_map_type_L0_skeleton(S1, reverse_imports_map)
        key_module, key_rel_type = compute_relative_type(key_type, current_module, reverse_imports_map)
        value_module, value_rel_type = compute_relative_type(value_type, current_module, reverse_imports_map)
        key_rel = reduce_rel_type( key_module, current_module, key_rel_type, reverse_imports_map )
        value_rel = reduce_rel_type( value_module, current_module, value_rel_type, reverse_imports_map )
        return current_module, f'smap[{key_rel}, {value_rel}]'
    
    
    if S1[0:5] == 'tupl[':
        typon_tuple_typon_type_list = read_tuple_types_typon_L0_skeleton(S1, reverse_imports_map)
        tuple_rel_types = []
        
        for arg_type in typon_tuple_typon_type_list:
            item_module, item_rel_type = compute_relative_type(arg_type, current_module, reverse_imports_map)
            item_rel = reduce_rel_type( item_module, current_module, item_rel_type, reverse_imports_map )
            tuple_rel_types.append(item_rel)
        
        S2 = ', '.join(tuple_rel_types)
        return current_module, f'tupl[{S2}]'
    
    
    if S1[0:4] == 'fxn[':
        arg_types_typon, return_type_typon = read_args_ret_from_fxn_typon_type(S1)
        arg_rel_types = []
        
        for arg_type in arg_types_typon:
            item_module, item_rel_type = compute_relative_type(arg_type, current_module, reverse_imports_map)
            item_rel = reduce_rel_type( item_module, current_module, item_rel_type, reverse_imports_map )
            arg_rel_types.append(item_rel)
        
        ret_module, ret_rel_type = compute_relative_type(return_type_typon, current_module, reverse_imports_map)
        ret_rel = reduce_rel_type( ret_module, current_module, ret_rel_type, reverse_imports_map )
        S2 = ', '.join(arg_rel_types)
        return current_module, f'fxn[[{S2}], {ret_rel}]'
    
    user_defined_typon_class_type = S1
    
    if user_defined_typon_class_type.find('|') != -1:
        target_module, type_name = user_defined_typon_class_type.split('|')
        return target_module, type_name
    
    i1 = user_defined_typon_class_type.find('.')
    
    if i1 == -1:
        return current_module, user_defined_typon_class_type
    
    raise Exception(f'could not parse type: {abs_type}' )


def resolve_map_type_L0_skeleton(typon_map_type, L0_skeleton):
    assert typon_map_type[0:5] in ['hmap[', 'smap[']
    S1 = typon_map_type.replace(' ', '')
    type_str = S1[5:-1]
    typon_typle = parsing_utils.read_delimited_types(type_str)
    assert len(typon_typle) == 2
    map_key_type, map_value_type = typon_typle
    return map_key_type, map_value_type


def read_tuple_types_typon_L0_skeleton(ws_removed_typon_tuple_type, L0_skeleton):
    ws_removed_typon_tuple_type = ws_removed_typon_tuple_type.replace(' ', '')
    interior_tuple_type_str = ws_removed_typon_tuple_type[5:-1]
    
    if len(interior_tuple_type_str) == 0:
        error_msgs = [
                    'invalid type declaration',
                    f'empty tuple not allowed: {ws_removed_typon_tuple_type}',
                ]
        exceptions.raise_exception_ue(error_msgs)
    
    typon_tuple_typon_type_list = parsing_utils.read_delimited_types(interior_tuple_type_str)
    return typon_tuple_typon_type_list


def read_args_ret_from_fxn_typon_type(ws_removed_typon_fxn_type):
    interior_fxn_type = ws_removed_typon_fxn_type[4:-1]
    
    if len(interior_fxn_type) == 0:
        error_msgs = [
                    'invalid type declaration',
                    f'empty fxn not allowed: {ws_removed_typon_fxn_type}',
                ]
        exceptions.raise_exception_ue(error_msgs)
    
    arg_closure_index = parsing_utils.get_char_closure(interior_fxn_type, 1)
    typon_fxn_arg_str = interior_fxn_type[1:arg_closure_index-1]
    arg_types_typon = parsing_utils.read_delimited_types(typon_fxn_arg_str)
    return_type_typon = interior_fxn_type[arg_closure_index+1:]
    return arg_types_typon, return_type_typon
